/* A generic RestLink middleware
 * Meant to be inserted in a server or in a middleware chain
 * can handle requests or dispatch them to next middleware object.
 * API derived from express.js / connect.js for least surprise.
 */
if (typeof define !== 'function') { var define = require('amdefine')(module); }

define(
[
	'underscore',
	'when',
	'extended-exceptions',
	'restlink/server/middleware/request_enrichments',
	'restlink/server/middleware/response_enrichments'
],
function(_, when, EE, RequestEnricher, ResponseEnricher) {
	"use strict";


	////////////////////////////////////
	var constants  = {};
	var defaults   = {};
	var exceptions = {};
	var methods    = {};


	////////////////////////////////////
	//constants. = ;


	////////////////////////////////////
	//defaults. = ;

	// default implementation, to be overriden or replaced (choice)
	// @param context : user-provided context object (passed along but untouched by the middleware. Can be left to undefined.
	// @param request : a request object that will be passed along. Can be anything.
	// @param response : a response object that will be passed along. Can be anything, will have been "enriched" with the middleware special methods.
	// @param next : a function to call to transfer to next middleware. Will throw if none.
	defaults.processing_function_ = function base_middleware_default_processing_function(context, request, response, next) {
		if(typeof this.next_middleware_ !== "undefined") {
			// pass to next
			next();
		}
		else {
			response.send();
		}
	};
	defaults.back_processing_function_ = undefined; // none by default or would cause overhead
	defaults.next_middleware_ = undefined; //< if chain, the middleware just after us to whom we may transfer the processing

	////////////////////////////////////
	//exceptions. = ;


	////////////////////////////////////
	//methods. = ;


	// add a chained middleware after this one.
	// will be inserted at the end of current chain if any.
	methods.use = function(middleware) {
		if(!this.next_middleware_) {
			this.next_middleware_ = middleware;
		}
		else {
			this.next_middleware_.use(middleware);
		}
	};

	function register_back_function(response, back_function, that) {
		var deferred = when.defer();
		response.middleware_.deferred_chain_.push(deferred);
		deferred.promise.spread(function(context, request, response) {
			back_function(context, request, response, that);
		});
	}

	// implementation of method "next" which is provided as a param
	// if a back function is given, it'll have to call send() again to continue the processing.
	function next_implementation(that, context, request, response, optional_back_function) {
		if(!that.next_middleware_) {
			// no middleware after us !
			// we should NOT have called next, then !
			throw new EE.IllegalStateError("Can't forward to next middleware, having none !");
		}
		else {
			// should optional replace the integrated or add to it ?
			// addition for now.
			if(typeof optional_back_function !== "undefined") {
				register_back_function(response, optional_back_function, that);
			}
		}

		that.next_middleware_.process_request_(context, request, response);
	}


	// wrapper around the user-defined processing function
	// @see processing_function_
	methods.process_request_ = function(context, request, response) {

		// create the "next" parameter
		var current_mw = this; // closure
		var next = function(optional_back_function) {
			// see above
			next_implementation(current_mw, context, request, response, optional_back_function);
		};

		// register the default back function if any
		if(typeof this.back_processing_function_ !== "undefined") {
			register_back_function(response, this.back_processing_function_, current_mw);
		}

		// call the (hopefully) user-defined processing function
		this.processing_function_(context, request, response, next, this);
	};


	// Launch the middleware processing chain.
	// should only be called on the head middleware !
	// Not mandatory if you wan to initialize the middleware manually
	methods.head_process_request = function(context, request, response) {

		// add some utility methods to the request
		RequestEnricher.process(request, context);

		// create the response if not created yet
		if(typeof response === "undefined") {
			response = this.prepare_response_for_processing_(request);
		}

		// initiate the middleware chain
		// and get the corresponding promise
		var promise = this.init_middleware_chain_on_this_response(response);

		// now use the usual function
		this.process_request_(context, request, response);

		return promise;
	};

	methods.prepare_response_for_processing_ = function(request) {
		var response = request.make_response(); // REM this response is set to "internal error" by default

		// we add some utility methods to the response
		ResponseEnricher.process(response, request, context);

		return response;
	};
	// initiate the middleware chain
	// by inserting a first deferred
	methods.init_middleware_chain_on_this_response = function(response) {
		// check params
		if(response.middleware_.deferred_chain_.length !== 0) {
			// WAT ? what is the use of calling this function
			// if everything is already done ??
			throw new EE.InvalidArgument("Offirmo Middleware : Middleware chain already initialized !");
		}

		// insert first deferred
		var deferred = when.defer();
		response.middleware_.deferred_chain_.push(deferred);

		return deferred.promise;
	};



	////////////////////////////////////
	Object.freeze(constants);
	Object.freeze(exceptions);
	Object.freeze(defaults);
	Object.freeze(methods);

	var DefinedClass = function RestlinkMiddlewareBase(process_func, process_back_func) {
		_.defaults( this, defaults );

		// other inits...
		if(typeof process_func !== "undefined") {
			this.processing_function_ = process_func; // replace
		}
		if(typeof process_back_func !== "undefined") {
			this.back_processing_function_ = process_back_func;
		}
	};

	DefinedClass.prototype.constants  = constants;
	DefinedClass.prototype.exceptions = exceptions;
	_.extend(DefinedClass.prototype, methods);


	////////////////////////////////////
	return {
		// objects are created via a factory, more future-proof
		'make_new'   : function(process_func, process_back_func) { return new DefinedClass(process_func, process_back_func); },
		// but we still expose the constructor to allow class inheritance
		'klass'      : DefinedClass,
		// exposing these allows convenient syntax and also prototypal inheritance
		'constants'  : constants,
		'exceptions' : exceptions,
		'defaults'   : defaults,
		'methods'    : methods
	};
}); // requirejs module
